# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

logger = logging.getLogger(__name__)

import pyomo.environ as pyo
from pyomo.contrib.alternative_solutions import aos_utils
from pyomo.contrib.alternative_solutions import Solution
from pyomo.contrib import appsi


def obbt_analysis(
    model,
    *,
    variables=None,
    rel_opt_gap=None,
    abs_opt_gap=None,
    refine_discrete_bounds=False,
    warmstart=True,
    solver="gurobi",
    solver_options={},
    tee=False,
):
    """
    Calculates the bounds on each variable by solving a series of min and max
    optimization problems where each variable is used as the objective function
    This can be applied to any class of problem supported by the selected
    solver.

    Parameters
    ----------
    model : ConcreteModel
        A concrete Pyomo model.
    variables: None or a collection of Pyomo _GeneralVarData variables
        The variables for which bounds will be generated. None indicates
        that all variables will be included. Alternatively, a collection of
        _GeneralVarData variables can be provided.
    rel_opt_gap : float or None
        The relative optimality gap for the original objective for which
        variable bounds will be found. None indicates that a relative gap
        constraint will not be added to the model.
    abs_opt_gap : float or None
        The absolute optimality gap for the original objective for which
        variable bounds will be found. None indicates that an absolute gap
        constraint will not be added to the model.
    refine_discrete_bounds : boolean
        Boolean indicating that new constraints should be added to the
        model at each iteration to tighten the bounds for discrete
        variables.
    warmstart : boolean
        Boolean indicating that the solver should be warmstarted from the
        best previously discovered solution.
    solver : string
        The solver to be used.
    solver_options : dict
        Solver option-value pairs to be passed to the solver.
    tee : boolean
        Boolean indicating that the solver output should be displayed.

    Returns
    -------
    variable_ranges
        A Pyomo ComponentMap containing the bounds for each variable.
        {variable: (lower_bound, upper_bound)}. An exception is raised when
        the solver encountered an issue.
    """
    bounds, solns = obbt_analysis_bounds_and_solutions(
        model,
        variables=variables,
        rel_opt_gap=rel_opt_gap,
        abs_opt_gap=abs_opt_gap,
        refine_discrete_bounds=refine_discrete_bounds,
        warmstart=warmstart,
        solver=solver,
        solver_options=solver_options,
        tee=tee,
    )
    return bounds


def obbt_analysis_bounds_and_solutions(
    model,
    *,
    variables=None,
    rel_opt_gap=None,
    abs_opt_gap=None,
    refine_discrete_bounds=False,
    warmstart=True,
    solver="gurobi",
    solver_options={},
    tee=False,
):
    """
    Calculates the bounds on each variable by solving a series of min and max
    optimization problems where each variable is used as the objective function
    This can be applied to any class of problem supported by the selected
    solver.

    Parameters
    ----------
    model : ConcreteModel
        A concrete Pyomo model.
    variables: None or a collection of Pyomo _GeneralVarData variables
        The variables for which bounds will be generated. None indicates
        that all variables will be included. Alternatively, a collection of
        _GeneralVarData variables can be provided.
    rel_opt_gap : float or None
        The relative optimality gap for the original objective for which
        variable bounds will be found. None indicates that a relative gap
        constraint will not be added to the model.
    abs_opt_gap : float or None
        The absolute optimality gap for the original objective for which
        variable bounds will be found. None indicates that an absolute gap
        constraint will not be added to the model.
    refine_discrete_bounds : boolean
        Boolean indicating that new constraints should be added to the
        model at each iteration to tighten the bounds for discrete
        variables.
    warmstart : boolean
        Boolean indicating that the solver should be warmstarted from the
        best previously discovered solution.
    solver : string
        The solver to be used.
    solver_options : dict
        Solver option-value pairs to be passed to the solver.
    tee : boolean
        Boolean indicating that the solver output should be displayed.

    Returns
    -------
    variable_ranges
        A Pyomo ComponentMap containing the bounds for each variable.
        {variable: (lower_bound, upper_bound)}. An exception is raised when
        the solver encountered an issue.
    solutions
        [Solution]
    """

    # TODO - parallelization

    logger.info("STARTING OBBT ANALYSIS")

    if warmstart:
        assert (
            variables == None
        ), "Cannot restrict variable list when warmstart is specified"
    all_variables = aos_utils.get_model_variables(model, include_fixed=False)
    if variables == None:
        variable_list = all_variables
    else:
        variable_list = list(variables)
    if warmstart:
        solutions = pyo.ComponentMap()
        for var in all_variables:
            solutions[var] = []

    num_vars = len(variable_list)
    logger.info(
        "Analyzing {} variables ({} total solves).".format(num_vars, 2 * num_vars)
    )
    orig_objective = aos_utils.get_active_objective(model)

    use_appsi = False
    if "appsi" in solver:
        opt = appsi.solvers.Gurobi()
        for parameter, value in solver_options.items():
            opt.gurobi_options[parameter] = value
        opt.config.stream_solver = tee
        opt.config.load_solution = False
        results = opt.solve(model)
        condition = results.termination_condition
        optimal_tc = appsi.base.TerminationCondition.optimal
        infeas_or_unbdd_tc = appsi.base.TerminationCondition.infeasibleOrUnbounded
        unbdd_tc = appsi.base.TerminationCondition.unbounded
        use_appsi = True
    else:
        opt = pyo.SolverFactory(solver)
        opt.available()
        for parameter, value in solver_options.items():
            opt.options[parameter] = value
        try:
            results = opt.solve(
                model, warmstart=warmstart, tee=tee, load_solutions=False
            )
        except ValueError:
            # An exception occurs if the solver does not recognize the warmstart option
            results = opt.solve(model, tee=tee, load_solutions=False)
        condition = results.solver.termination_condition
        optimal_tc = pyo.TerminationCondition.optimal
        infeas_or_unbdd_tc = pyo.TerminationCondition.infeasibleOrUnbounded
        unbdd_tc = pyo.TerminationCondition.unbounded
    logger.info("Performing initial solve of model.")

    if condition != optimal_tc:
        raise RuntimeError(
            ("OBBT cannot be applied, " "TerminationCondition = {}").format(
                condition.value
            )
        )
    if use_appsi:
        results.solution_loader.load_vars(solution_number=0)
    else:
        model.solutions.load_from(results)
    if warmstart:
        _add_solution(solutions)
    orig_objective_value = pyo.value(orig_objective)
    logger.info("Found optimal solution, value = {}.".format(orig_objective_value))
    aos_block = aos_utils._add_aos_block(model, name="_obbt")
    # placeholder for objective
    aos_block.var_objective = pyo.Objective(expr=0)
    logger.info("Added block {} to the model.".format(aos_block))
    obj_constraints = aos_utils._add_objective_constraint(
        aos_block, orig_objective, orig_objective_value, rel_opt_gap, abs_opt_gap
    )
    if refine_discrete_bounds:
        aos_block.bound_constraints = pyo.ConstraintList()
    new_constraint = False
    if len(obj_constraints) > 0:
        new_constraint = True
    orig_objective.deactivate()

    if use_appsi:
        opt.update_config.check_for_new_or_removed_constraints = new_constraint
        opt.update_config.check_for_new_or_removed_vars = False
        opt.update_config.check_for_new_or_removed_params = False
        opt.update_config.check_for_new_objective = True
        opt.update_config.update_constraints = False
        opt.update_config.update_vars = False
        opt.update_config.update_params = False
        opt.update_config.update_named_expressions = False
        opt.update_config.update_objective = True
        opt.update_config.treat_fixed_vars_as_params = False

    variable_bounds = pyo.ComponentMap()
    solns = [Solution(model, all_variables, objective=orig_objective)]

    senses = [(pyo.minimize, "LB"), (pyo.maximize, "UB")]

    iteration = 1
    total_iterations = len(senses) * num_vars
    for idx in range(len(senses)):
        sense = senses[idx][0]
        bound_dir = senses[idx][1]

        for var in variable_list:
            if idx == 0:
                variable_bounds[var] = [None, None]

            aos_block.var_objective.expr = var
            aos_block.var_objective.sense = sense

            if warmstart:
                _update_values(var, bound_dir, solutions)

            if use_appsi:
                opt.update_config.check_for_new_or_removed_constraints = new_constraint
            if use_appsi:
                opt.config.stream_solver = tee
                results = opt.solve(model)
                condition = results.termination_condition
            else:
                try:
                    results = opt.solve(
                        model, warmstart=warmstart, tee=tee, load_solutions=False
                    )
                except ValueError:
                    # An exception occurs if the solver does not recognize the warmstart option
                    results = opt.solve(model, tee=tee, load_solutions=False)
                condition = results.solver.termination_condition
            new_constraint = False

            if condition == optimal_tc:
                if use_appsi:
                    results.solution_loader.load_vars(solution_number=0)
                else:
                    model.solutions.load_from(results)
                solns.append(Solution(model, all_variables, objective=orig_objective))

                if warmstart:
                    _add_solution(solutions)
                obj_val = pyo.value(var)
                variable_bounds[var][idx] = obj_val

                if refine_discrete_bounds and not var.is_continuous():
                    if sense == pyo.minimize and var.lb < obj_val:
                        aos_block.bound_constraints.add(var >= obj_val)
                        new_constraint = True

                    if sense == pyo.maximize and var.ub > obj_val:
                        aos_block.bound_constraints.add(var <= obj_val)
                        new_constraint = True

            # An infeasibleOrUnbounded status code will imply the problem is
            # unbounded since feasibility has been established previously
            elif condition == infeas_or_unbdd_tc or condition == unbdd_tc:
                if sense == pyo.minimize:
                    variable_bounds[var][idx] = float("-inf")
                else:
                    variable_bounds[var][idx] = float("inf")
            else:  # pragma: no cover
                logger.warn(
                    (
                        "Unexpected condition for the variable {} {} problem."
                        "TerminationCondition = {}"
                    ).format(var.name, bound_dir, condition.value)
                )

            var_value = variable_bounds[var][idx]
            logger.info(
                "Iteration {}/{}: {}_{} = {}".format(
                    iteration, total_iterations, var.name, bound_dir, var_value
                )
            )

            if idx == 1:
                variable_bounds[var] = tuple(variable_bounds[var])

            iteration += 1

    aos_block.deactivate()
    orig_objective.activate()

    logger.info("COMPLETED OBBT ANALYSIS")

    return variable_bounds, solns


def _add_solution(solutions):
    """Add the current variable values to the solution list."""
    for var in solutions:
        solutions[var].append(pyo.value(var))


def _update_values(var, bound_dir, solutions):
    """
    Set the values of all variables to the best solution seen previously for
    the current objective function.
    """
    if bound_dir == "LB":
        value = min(solutions[var])
    else:
        value = max(solutions[var])
    idx = solutions[var].index(value)
    for variable in solutions:
        variable.set_value(solutions[variable][idx])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import re

"""This script is run using the Gurobi/system python. Do not assume any
third party packages are available!

"""
from gurobipy import gurobi, read, GRB
import sys

if sys.version_info[0] < 3:
    from itertools import izip as zip

GUROBI_VERSION = gurobi.version()

# NOTE: this function / module is independent of Pyomo, and only relies on the
#       GUROBI python bindings. consequently, nothing in this function should
#       throw an exception that is expected to be handled by Pyomo - it won't be.
#       rather, print an error message and return - the caller will know to look
#       in the logs in case of a failure.


def _is_numeric(x):
    try:
        float(x)
    except ValueError:
        return False
    return True


def gurobi_run(model_file, warmstart_file, mipgap, options, suffixes):
    # figure out what suffixes we need to extract.
    extract_duals = False
    extract_slacks = False
    extract_reduced_costs = False
    for suffix in suffixes:
        flag = False
        if re.match(suffix, "dual"):
            extract_duals = True
            flag = True
        if re.match(suffix, "slack"):
            extract_slacks = True
            flag = True
        if re.match(suffix, "rc"):
            extract_reduced_costs = True
            flag = True
        if not flag:
            print(
                "***The GUROBI solver plugin cannot extract solution suffix=" + suffix
            )
            return

    # Load the lp model
    model = read(model_file)

    # if the use wants to extract duals or reduced costs and the
    # model has quadratic constraints then we need to set the
    # QCPDual param to 1 (which apparently makes the solve more
    # expensive in the quadratic case). If we do not set this param
    # and and we attempt to access these suffixes in the solution
    # printing the module will crash (when we have a QCP)
    if GUROBI_VERSION[0] >= 5:
        if (extract_reduced_costs is True) or (extract_duals is True):
            model.setParam(GRB.Param.QCPDual, 1)

    if model is None:
        print(
            "***The GUROBI solver plugin failed to load the input LP file=" + model_file
        )
        return

    if warmstart_file is not None:
        model.read(warmstart_file)

    # set the mipgap if specified.
    if mipgap is not None:
        model.setParam("MIPGap", mipgap)

    # set all other solver parameters, if specified.
    # GUROBI doesn't throw an exception if an unknown
    # key is specified, so you have to stare at the
    # output to see if it was accepted.
    for key, value in options.items():
        # When options come from the pyomo command, all
        # values are string types, so we try to cast
        # them to a numeric value in the event that
        # setting the parameter fails.
        try:
            model.setParam(key, value)
        except TypeError:
            # we place the exception handling for checking
            # the cast of value to a float in another
            # function so that we can simply call raise here
            # instead of except TypeError as e / raise e,
            # because the latter does not preserve the
            # Gurobi stack trace
            if not _is_numeric(value):
                model.close()
                raise
            model.setParam(key, float(value))

    if 'relax_integrality' in options:
        for v in model.getVars():
            if v.vType != GRB.CONTINUOUS:
                v.vType = GRB.CONTINUOUS
        model.update()

    # optimize the model
    model.optimize()

    # This attribute needs to be extracted before
    # calling model.getVars() or model.getConstrs().
    # Apparently it gets reset by those methods.
    wall_time = model.getAttr(GRB.Attr.Runtime)

    solver_status = model.getAttr(GRB.Attr.Status)
    solution_status = None
    return_code = 0
    if solver_status == GRB.LOADED:
        status = 'aborted'
        message = 'Model is loaded, but no solution information is available.'
        term_cond = 'error'
        solution_status = 'unknown'
    elif solver_status == GRB.OPTIMAL:
        status = 'ok'
        message = 'Model was solved to optimality (subject to tolerances), and an optimal solution is available.'
        term_cond = 'optimal'
        solution_status = 'optimal'
    elif solver_status == GRB.INFEASIBLE:
        status = 'warning'
        message = 'Model was proven to be infeasible.'
        term_cond = 'infeasible'
        solution_status = 'infeasible'
    elif solver_status == GRB.INF_OR_UNBD:
        status = 'warning'
        message = 'Problem proven to be infeasible or unbounded.'
        term_cond = 'infeasibleOrUnbounded'
        solution_status = 'unsure'
    elif solver_status == GRB.UNBOUNDED:
        status = 'warning'
        message = 'Model was proven to be unbounded.'
        term_cond = 'unbounded'
        solution_status = 'unbounded'
    elif solver_status == GRB.CUTOFF:
        status = 'aborted'
        message = 'Optimal objective for model was proven to be worse than the value specified in the Cutoff  parameter. No solution information is available.'
        term_cond = 'minFunctionValue'
        solution_status = 'unknown'
    elif solver_status == GRB.ITERATION_LIMIT:
        status = 'aborted'
        message = 'Optimization terminated because the total number of simplex iterations performed exceeded the value specified in the IterationLimit parameter.'
        term_cond = 'maxIterations'
        solution_status = 'stoppedByLimit'
    elif solver_status == GRB.NODE_LIMIT:
        status = 'aborted'
        message = 'Optimization terminated because the total number of branch-and-cut nodes explored exceeded the value specified in the NodeLimit parameter.'
        term_cond = 'maxEvaluations'
        solution_status = 'stoppedByLimit'
    elif solver_status == GRB.TIME_LIMIT:
        status = 'aborted'
        message = 'Optimization terminated because the time expended exceeded the value specified in the TimeLimit parameter.'
        term_cond = 'maxTimeLimit'
        solution_status = 'stoppedByLimit'
    elif hasattr(GRB, "WORK_LIMIT") and (solver_status == GRB.WORK_LIMIT):
        status = 'aborted'
        message = 'Optimization terminated because the work expended exceeded the value specified in the WorkLimit parameter.'
        term_cond = 'maxTimeLimit'
        solution_status = 'stoppedByLimit'
    elif solver_status == GRB.SOLUTION_LIMIT:
        status = 'aborted'
        message = 'Optimization terminated because the number of solutions found reached the value specified in the SolutionLimit parameter.'
        term_cond = 'stoppedByLimit'
        solution_status = 'stoppedByLimit'
    elif solver_status == GRB.INTERRUPTED:
        status = 'aborted'
        message = 'Optimization was terminated by the user.'
        term_cond = 'error'
        solution_status = 'error'
    elif solver_status == GRB.NUMERIC:
        status = 'error'
        message = (
            'Optimization was terminated due to unrecoverable numerical difficulties.'
        )
        term_cond = 'error'
        solution_status = 'error'
    elif solver_status == GRB.SUBOPTIMAL:
        status = 'warning'
        message = 'Unable to satisfy optimality tolerances; a sub-optimal solution is available.'
        term_cond = 'other'
        solution_status = 'feasible'
    # note that USER_OBJ_LIMIT was added in Gurobi 7.0, so it may not be present
    elif (solver_status is not None) and (
        solver_status == getattr(GRB, 'USER_OBJ_LIMIT', None)
    ):
        status = 'aborted'
        message = (
            "User specified an objective limit "
            "(a bound on either the best objective "
            "or the best bound), and that limit has "
            "been reached. Solution is available."
        )
        term_cond = 'other'
        solution_status = 'stoppedByLimit'
    else:
        status = 'error'
        message = "Unhandled Gurobi solve status (" + str(solver_status) + ")"
        term_cond = 'error'
        solution_status = 'error'
    assert solution_status is not None

    sense = model.getAttr(GRB.Attr.ModelSense)
    try:
        obj_value = model.getAttr(GRB.Attr.ObjVal)
    except:
        obj_value = None
        if term_cond == "unbounded":
            if sense < 0:
                # maximize
                obj_value = float('inf')
            else:
                # minimize
                obj_value = float('-inf')
        elif term_cond == "infeasible":
            if sense < 0:
                # maximize
                obj_value = float('-inf')
            else:
                # minimize
                obj_value = float('inf')

    result = {}
    problem = result['problem'] = {}
    problem['name'] = model.getAttr(GRB.Attr.ModelName)

    # TODO: find out about bounds and fix this with error checking
    # this line fails for some reason so set the value to unknown
    try:
        bound = model.getAttr(GRB.Attr.ObjBound)
    except Exception:
        if term_cond == 'optimal':
            bound = obj_value
        else:
            bound = None

    if sense < 0:
        problem["sense"] = "maximize"
        if bound is None:
            bound = float('inf')
        problem["upper_bound"] = bound
    else:
        problem["sense"] = "minimize"
        if bound is None:
            bound = float('-inf')
        problem["lower_bound"] = bound

    # TODO: Get the number of objective functions from GUROBI
    n_objs = 1
    problem["number_of_objectives"] = n_objs

    cons = model.getConstrs()
    qcons = []
    if GUROBI_VERSION[0] >= 5:
        qcons = model.getQConstrs()
    problem["number_of_constraints"] = len(cons) + len(qcons) + model.NumSOS

    vars = model.getVars()
    problem["number_of_variables"] = len(vars)

    n_binvars = model.getAttr(GRB.Attr.NumBinVars)
    problem["number_of_binary_variables"] = n_binvars

    n_intvars = model.getAttr(GRB.Attr.NumIntVars)
    problem["number_of_integer_variables"] = n_intvars
    problem["number_of_continuous_variables"] = len(vars) - n_intvars
    problem["number_of_nonzeros"] = model.getAttr(GRB.Attr.NumNZs)

    # write out the information required by results.solver
    solver = result['solver'] = {}

    solver['status'] = status
    solver['return_code'] = return_code
    solver['message'] = message
    solver['wall_time'] = wall_time
    solver['termination_condition'] = term_cond
    solver['termination_message'] = message

    is_discrete = False
    if model.getAttr(GRB.Attr.IsMIP):
        is_discrete = True

    if (term_cond == 'optimal') or (model.getAttr(GRB.Attr.SolCount) >= 1):
        solution = result['solution'] = {}
        solution['status'] = solution_status
        solution['message'] = message
        solution['objective'] = obj_value
        solution['gap'] = 0.0

        vals = model.getAttr("X", vars)
        names = model.getAttr("VarName", vars)
        solution['var'] = {name: val for name, val in zip(names, vals)}

        if extract_reduced_costs and not is_discrete:
            vals = model.getAttr("Rc", vars)
            solution['varrc'] = {name: val for name, val in zip(names, vals)}

        if extract_duals or extract_slacks:
            con_names = model.getAttr("ConstrName", cons)
            if GUROBI_VERSION[0] >= 5:
                qcon_names = model.getAttr("QCName", qcons)

        if extract_duals and not is_discrete:
            # Pi attributes in Gurobi are the constraint duals
            vals = model.getAttr("Pi", cons)
            solution['constraintdual'] = {
                name: val for name, val in zip(con_names, vals)
            }
            if GUROBI_VERSION[0] >= 5:
                # QCPI attributes in Gurobi are the constraint duals
                vals = model.getAttr("QCPi", qcons)
                solution['constraintdual'].update(zip(qcon_names, vals))

        if extract_slacks:
            vals = model.getAttr("Slack", cons)
            solution['constraintslack'] = {
                name: val for name, val in zip(con_names, vals)
            }
            if GUROBI_VERSION[0] >= 5:
                vals = model.getAttr("QCSlack", qcons)
                solution['constraintslack'].update(zip(qcon_names, vals))

    model.close()
    model = None
    return result


def write_result(result, soln_file):
    with open(soln_file, "w+") as FILE:
        for section, data in result.items():
            FILE.write(f'section:{section}\n')
            for key, val in data.items():
                if val.__class__ is dict:
                    for name, v in val.items():
                        FILE.write(f'{key}:{name}:{v}\n')
                else:
                    FILE.write(f'{key}:{val}\n')



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import operator
import itertools
import pyomo.core.base.var
import pyomo.core.base.constraint
from pyomo.core import is_fixed, value
from pyomo.core.base.PyomoModel import ConcreteModel
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.var import Var
from pyomo.solvers.plugins.solvers.mosek_direct import MOSEKDirect
from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver
from pyomo.solvers.plugins.solvers.direct_solver import DirectSolver
from pyomo.solvers.plugins.solvers.direct_or_persistent_solver import (
    DirectOrPersistentSolver,
)
from pyomo.opt.base import SolverFactory
from pyomo.core.kernel.conic import _ConicBase
from pyomo.core.kernel.block import block


@SolverFactory.register('mosek_persistent', doc='Persistent python interface to MOSEK.')
class MOSEKPersistent(PersistentSolver, MOSEKDirect):
    """
    This class provides a persistent interface between pyomo and MOSEK's Optimizer API.
    As a child to the MOSEKDirect class, this interface does not need any file IO.
    Furthermore, the persistent interface preserves the MOSEK task object, allowing
    users to make incremental changes (such as removing variables/constraints, modifying
    variables, adding columns etc.) to their models. Note that users are responsible for
    informing the persistent interface of any incremental change. For instance, if a new
    variable is defined, then it would need to be added explicitly by calling the add_var
    method, before the solver knows of its existence.
    Keyword Arguments
    -----------------
    type: str
        String indicating the class type of the solver instance.
    name: str
        String representing either the class type of the solver instance or an assigned name.
    doc: str
        Documentation for the solver
    options: dict
        Dictionary of solver options
    """

    def __init__(self, **kwds):
        kwds.setdefault('type', 'mosek_persistent')
        MOSEKDirect.__init__(self, **kwds)

        self._pyomo_model = kwds.pop('model', None)
        if self._pyomo_model is not None:
            self.set_instance(self._pyomo_model, **kwds)

    def _warm_start(self):
        MOSEKDirect._warm_start(self)

    def add_vars(self, var_seq):
        """
        Add multiple variables to the MOSEK task object in one method call.

        This will keep any existing model components intact.

        Parameters
        ----------
        var_seq: tuple/list of Var
        """
        self._add_vars(var_seq)

    def add_constraints(self, con_seq):
        """
        Add multiple constraints to the MOSEK task object in one method call.

        This will keep any existing model components intact.

        NOTE: If this method is used to add cones, then the cones should be
        passed as constraints. Use the add_block method for conic_domains.

        Parameters
        ----------
        con_seq: tuple/list of Constraint (scalar Constraint or single ConstraintData)
        """
        self._add_constraints(con_seq)

    def remove_var(self, solver_var):
        """
        Remove a single variable from the model as well as the MOSEK task.
        This will keep any other model components intact.
        Parameters
        ----------
        solver_var: Var (scalar Var or single VarData)
        """
        self.remove_vars(solver_var)

    def remove_vars(self, *solver_vars):
        """
        Remove multiple scalar variables from the model as well as the MOSEK task. The
        user can pass an unpacked list of scalar variables.
        This will keep any other model components intact.
        Parameters
        ----------
        *solver_var: Var (scalar Var or single VarData)
        """
        try:
            var_ids = []
            for v in solver_vars:
                var_ids.append(self._pyomo_var_to_solver_var_map[v])
                self._symbol_map.removeSymbol(v)
                del self._referenced_variables[v]
                del self._pyomo_var_to_solver_var_map[v]
            self._solver_model.removevars(var_ids)
        except KeyError:
            v_name = self._symbol_map.getSymbol(v, self._labeler)
            raise ValueError(
                "Variable {} needs to be added before removal.".format(v_name)
            )
        var_num = self._solver_model.getnumvar()
        for i, v in enumerate(self._pyomo_var_to_solver_var_map):
            self._pyomo_var_to_solver_var_map[v] = i
        self._solver_var_to_pyomo_var_map = dict(
            zip((range(var_num)), self._pyomo_var_to_solver_var_map.keys())
        )

    def remove_constraint(self, solver_con):
        """
        Remove a single constraint from the model as well as the MOSEK task.

        This will keep any other model components intact.

        To remove a conic-domain, you should use the remove_block method.
        Parameters
        ----------
        solver_con: Constraint (scalar Constraint or single ConstraintData)
        """
        self.remove_constraints(solver_con)

    def remove_constraints(self, *solver_cons):
        """
        Remove multiple constraints from the model as well as the MOSEK task in one
        method call.

        This will keep any other model components intact.
        To remove conic-domains, use the remove_block method.

        Parameters
        ----------
        *solver_cons: Constraint (scalar Constraint or single ConstraintData)
        """
        lq_cons = tuple(
            itertools.filterfalse(lambda x: isinstance(x, _ConicBase), solver_cons)
        )
        cone_cons = tuple(filter(lambda x: isinstance(x, _ConicBase), solver_cons))
        try:
            lq = []
            cones = []
            for c in lq_cons:
                lq.append(self._pyomo_con_to_solver_con_map[c])
                self._symbol_map.removeSymbol(c)
                del self._pyomo_con_to_solver_con_map[c]
            for c in cone_cons:
                cones.append(self._pyomo_cone_to_solver_cone_map[c])
                self._symbol_map.removeSymbol(c)
                del self._pyomo_cone_to_solver_cone_map[c]
            self._solver_model.removecons(lq)
            self._solver_model.removecones(cones)
            lq_num = self._solver_model.getnumcon()
            cone_num = self._solver_model.getnumcone()
        except KeyError:
            c_name = self._symbol_map.getSymbol(c, self._labeler)
            raise ValueError(
                "Constraint/Cone {} needs to be added before removal.".format(c_name)
            )
        self._solver_con_to_pyomo_con_map = dict(
            zip(range(lq_num), self._pyomo_con_to_solver_con_map.keys())
        )
        self._solver_cone_to_pyomo_cone_map = dict(
            zip(range(cone_num), self._pyomo_cone_to_solver_cone_map.keys())
        )
        for i, c in enumerate(self._pyomo_con_to_solver_con_map):
            self._pyomo_con_to_solver_con_map[c] = i
        for i, c in enumerate(self._pyomo_cone_to_solver_cone_map):
            self._pyomo_cone_to_solver_cone_map[c] = i

    def update_var(self, solver_var):
        """
        Update a single variable in solver model. This method allows fixing/unfixing,
        changing variable type and bounds.
        Parameters
        ----------
        solver_var: Var
        """

        self.update_vars(solver_var)

    def update_vars(self, *solver_vars):
        """
        Update multiple scalar variables in solver model. This method allows fixing/unfixing,
        changing variable types and bounds.
        Parameters
        ----------
        *solver_var: Constraint (scalar Constraint or single ConstraintData)
        """
        try:
            var_ids = []
            for v in solver_vars:
                var_ids.append(self._pyomo_var_to_solver_var_map[v])
            vtypes = tuple(map(self._mosek_vartype_from_var, solver_vars))
            lbs = tuple(
                (
                    value(v)
                    if v.fixed
                    else -float('inf') if value(v.lb) is None else value(v.lb)
                )
                for v in solver_vars
            )
            ubs = tuple(
                (
                    value(v)
                    if v.fixed
                    else float('inf') if value(v.ub) is None else value(v.ub)
                )
                for v in solver_vars
            )
            fxs = tuple(v.is_fixed() for v in solver_vars)
            bound_types = tuple(map(self._mosek_bounds, lbs, ubs, fxs))
            self._solver_model.putvartypelist(var_ids, vtypes)
            self._solver_model.putvarboundlist(var_ids, bound_types, lbs, ubs)
        except KeyError:
            v_name = self._symbol_map.getSymbol(v, self._labeler)
            raise ValueError(
                "Variable {} needs to be added before it can be modified.".format(
                    v_name
                )
            )

    def _add_column(self, var, obj_coef, constraints, coefficients):
        self.add_var(var)
        var_num = self._solver_model.getnumvar()
        self._solver_model.putcj(var_num - 1, obj_coef)
        self._solver_model.putacol(var_num - 1, constraints, coefficients)
        self._referenced_variables[var] = len(constraints)

    def write(self, filename):
        """
        Write the model to a file. MOSEK can write files in various
        popular formats such as: lp, mps, ptf, cbf etc.
        In addition to the file formats mentioned above, MOSEK can
        also write files to native formats such as : opf, task and
        jtask. The task format is binary, and is the preferred format
        for sharing with the MOSEK staff in case of queries, since it saves
        the status of the problem and the solver down the smallest detail.
        Parameters:
        filename: str (Name of the output file, including the desired extension)
        """
        if 'task' in filename.split("."):
            self._solver_model.writetask(filename)
        else:
            self._solver_model.writedata(filename)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import numpy as np, pandas as pd
from os.path import join, abspath, dirname
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.reactor_design.reactor_design import (
    ReactorDesignExperiment,
)

import pyomo.environ as pyo
from pyomo.contrib.parmest.utils.model_utils import update_model_from_suffix


def main():
    # Read in file
    # Read in data
    file_dirname = dirname(abspath(str(__file__)))
    file_name = abspath(join(file_dirname, "reactor_data.csv"))
    data = pd.read_csv(file_name)

    experiment = ReactorDesignExperiment(data, 0)

    # Call the experiment's model using get_labeled_model
    reactor_model = experiment.get_labeled_model()

    example_suffix = "unknown_parameters"
    suffix_obj = reactor_model.unknown_parameters
    var_list = list(suffix_obj.keys())  # components
    orig_var_vals = np.array([pyo.value(v) for v in var_list])  # numeric var values

    # Print original values
    print("Original sigma values")
    print("----------------------")
    print(orig_var_vals)

    # Update the suffix with new values
    new_vals = orig_var_vals + 0.5

    print("New sigma values")
    print("----------------")
    print(new_vals)

    # Here we are updating the values of the unknown parameters
    # You must know the length of the list and order of the suffix items to update them correctly
    update_model_from_suffix(suffix_obj, new_vals)

    # Updated values
    print("Updated sigma values :")
    print("-----------------------")
    new_var_vals = np.array([pyo.value(v) for v in var_list])
    print(new_var_vals)

    # Return the suffix obj, original and new values for further use if needed
    return suffix_obj, new_vals, new_var_vals


if __name__ == "__main__":
    main()
